// XPCOM Objects
const gPrefBranch = prefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);

const gDefaults = {generator: "RSS Editor (http://rsseditor.mozdev.org)", docs: "http://blogs.law.harvard.edu/tech/rss"};

// prototypes
function sdiObject()
{
	this.baseTitle = document.documentElement.getAttribute("title");
	// the target can be the current document or the document on the top of the frameset (in practice, window.top is used even if no frameset)
	this.target = window.top ? window.top.document : document;
	// set the title whether the window is initialized (document.title) or not (setAttribute)
	this.target.documentElement.setAttribute("title", this.target.title = this.name + " - " + this.baseTitle);
}
sdiObject.prototype =
{
	_content : null,
	get content() { return this._content; },
	set content(x) { this._content = x; this.updateTitle(); return x; },

	_modified : false,
	get modified() { return this._modified; },
	set modified(x) { this._modified = x; this.updateTitle(); return x; },

	get name()
	{
		if (this._content === null)
			return localize("untitled");
		else if ("leafName" in this._content)
			return this._content.leafName;
		else if ("spec" in this._content)
		{
			var tmpURI = this._content.clone();
			try { tmpURI.password = ""; } catch(e) {} // ignore errors
			return tmpURI.spec;
		}
		else
			return localize("### ERROR ###"); // return a optionally localized error message if everything fails
	},

	updateTitle : function()
	{
		var mmark = this._modified ? " * " : " - ";
		this.target.title = this.name + mmark + this.baseTitle; // document.title works in tab and window, window.title works in window only
	}
}


// Must not use prototype because it contains objects
function RSSChannel()
{
	this.title = "";
	this.link = "";
	this.description = "";
	this.language = "";
	this.copyright = "";
	this.managingEditor = "";
	this.webMaster = "";
	this.pubDate = "";
	this.lastBuildDate = "";
	this.category = "";
	this.domain = "";
	this.generator = "";
	this.docs = "";
	this.cloud = {domain:"", port:"", path:"", registerProcedure:"", protocol:""};
	this.ttl = "";
	this.image = {url: "", title:"", link:"", description:"", width:"", height:""};
	this.rating = "";
	this.textInput = {title:"", description:"", name:"", link:""};
	this.skipHours =  [];
	this.skipDays =  [];
	this.stylesheet = {href:"", type:""};
}


// RSSItem must not use a prototype to make it serializable by toSource()
function RSSItem()
{
	this.title = "";
	this.link = "";
	this.desc = "";
	this.author = "";
	this.category = "";
	this.domain = ""; // domain attr in <category>
	this.comments = "";
	this.enclosure = {url: "", length: "", type: ""};
	this.guid = "";
	this.isPermaLink = false; // isPermaLink attr in guid
	this.pubDate = "";
	this.source = "";
	this.sourceurl = ""; // url attr in <source>
}


// Objects (static)
const controller =
{
	supportsCommand : function(cmd)
	{
		switch (cmd)
		{
			case "cmd_cut":
			case "cmd_copy":
			case "cmd_paste":
			case "cmd_delete":
				return true;
			default:
				return false;
		}
	},

	isCommandEnabled : function(cmd)
	{
		var e1 = document.getElementById("tabItems").selected;
		var tree = document.getElementById("tree");
		var selList = treeSelection(tree);
		var e2 =  e1 && selList.length>0;
		switch (cmd)
		{
			case "cmd_delete":
			case "cmd_cut":
			case "cmd_copy":
				return e2;
			case "cmd_paste":
				return e1;
			default:
				return false;
		}
	},

	doCommand : function(cmd)
	{
		switch (cmd)
		{
			case "cmd_cut":
			case "cmd_copy":
				itemsCopy();
				if (cmd == "cmd_copy") break;
				itemsDelete(false);
				break;
			case "cmd_delete":
				itemsDelete(true);
				break;
			case "cmd_paste":
				itemsPaste();
				break;
		}
	},

	onEvent : function(evt) {}
};


// Public variables
var itemSelectDisabled = false; // temporary disable event for manipulating selected row in list
var updateItemCmdsDisabled = false; // temporary disable update of item commands
var selectedItem = null; // selected item for editing
var channelData = new RSSChannel(); // channel data;
var subject1;

// Listen to pref updates
var gPrefListener =
{
	observe: function(subject, topic, prefName)
	{
		if (topic != "nsPref:changed") return;
		switch (prefName)
		{
			case "smallIcons":
				if (boolPref(prefName))
					document.getElementById("rsseditor_toolbar").setAttribute("iconsize", "small");
				else
					document.getElementById("rsseditor_toolbar").removeAttribute("iconsize");

			case "guidReadOnly":
				if (boolPref(prefName))
					document.getElementById("itemGuid").setAttribute("readonly", "true");
				else
					document.getElementById("itemGuid").removeAttribute("readonly");

			case "UTCTime":
				var list = document.getElementById("itemList");
				var n = list.childNodes.length;
				for (var i=0; i<n; i++)
					list.childNodes[i].firstChild.childNodes[1].setAttribute("label", time(list.childNodes[i].firstChild.childNodes[1].getAttribute("label")));
			// Todo: update channel/pubdate, channel/lastbuilddate, item/pubdate to new time format

			default:
				if (prefName.substring(0, 5) == "show.")
				{
					var element = document.getElementById(prefName);
					if (element)
					{
						if (boolPref(prefName))
							element.removeAttribute("hidden");
						else
							element.setAttribute("hidden", "true");
					}
				}
		}
	}
}


// Date Updates
function channelLastBuildDateUpdate()
{
	boolPref("lastBuildDateAuto") && setValue("channelLastBuildDate", time());
	// Mark document modified
	sdi.modified = true;
}

function dateUpdates()
{
	if (boolPref("itemPubDateAuto"))
	{
		setValue("itemPubDate", time());
		itemInput(document.getElementById("itemPubDate"));
		// channelLastBuildDateUpdate is done in itemInput
	}
	else
	{
		channelLastBuildDateUpdate();
	}
}


// Reset channel fields
function resetChannel()
{
	setValue("channelTitle", channelData.title);
	setValue("channelLink", channelData.link);
	setValue("channelDesc", channelData.description);
	setValue("channelLang", channelData.language);
	setValue("channelCopyright", channelData.copyright);
	setValue("channelManagingEditor", channelData.managingEditor);
	setValue("channelWebMaster", channelData.webMaster);
	setValue("channelPubDate", channelData.pubDate);
	setValue("channelLastBuildDate", channelData.lastBuildDate);
	setValue("channelCategory", channelData.category);
	setValue("channelDomain", channelData.domain);
	setValue("channelGenerator", channelData.generator);
	setValue("channelDocs", channelData.docs);
	setValue("channelCloudDomain", channelData.cloud.domain);
	setValue("channelCloudPort", channelData.cloud.port);
	setValue("channelCloudPath", channelData.cloud.path);
	setValue("channelCloudRegisterProcedure", channelData.cloud.registerProcedure);
	setValue("channelCloudProtocol", channelData.cloud.protocol);
	setValue("channelTtl", channelData.ttl);
	setValue("channelRating", channelData.rating);
	setValue("channelImage", channelData.image.url);
	setValue("channelTextInputTitle", channelData.textInput.title);
	setValue("channelTextInputDescription", channelData.textInput.description);
	setValue("channelTextInputName", channelData.textInput.name);
	setValue("channelTextInputLink", channelData.textInput.link);

	var i, a;
	for (i=0; i<24; i++)
		document.getElementById("channelSkipHours." + i).checked = channelData.skipHours[i];
	for (i=0; i<7; i++)
		document.getElementById("channelSkipDays." + i).checked = channelData.skipDays[i];

	setValue("miscStylesheetHref", channelData.stylesheet.href);
	setValue("miscStylesheetType", channelData.stylesheet.type);
}


// Enable/disable commands for items
function updateItemCmds()
{
	if (updateItemCmdsDisabled) return;

	var e1 = document.getElementById("tabItems").selected;
	var tree = document.getElementById("tree");
	var selList = treeSelection(tree);
	var e2 =  e1 && selList.length>0;

	var a = {
		importCmd: e1,
		itemInsertCmd: e1,
		itemAppendCmd: e1,
		itemDeleteCmd: e2,
		itemRevertCmd: e1 && (selectedItem != null),
		itemMoveUpCmd: e2 && selList[0].start > 0,
		itemMoveDownCmd: e2 && selList[selList.length - 1].end < tree.view.rowCount - 1
	};
	for (var i in a)
	{
		if (!a[i])
			document.getElementById(i).setAttribute("disabled", "true");
		else
			document.getElementById(i).removeAttribute("disabled");
	}
}


// Return tree selection, simplified and sorted
function treeSelection(tree)
{
	var sel = tree.view.selection;
	var i, j, count;
	var selList = [], selS = [], selE = [];
	var start = {}, end = {};

	// sorting starts and ends individually - e.g. change 2-7, 3-6 to 2-6, 3-7
	count = sel.getRangeCount();
	for (i=0; i<count; i++)
	{
		sel.getRangeAt(i, start, end);
		// always push the smaller value to selS, and the bigger to selE
		if (start.value <= end.value)
		{
			selS.push(start.value);
			selE.push(end.value);
		}
		else
		{
			selS.push(end.value);
			selE.push(start.value);
		}
	}

	selS.sort(numericCompare); selE.sort(numericCompare);

	count = selS.length;
	for (i=0; i<count; i=j)
	{
		// only the first "start" and the last "end" is needed if the blocks are connected or overlapping
		for (j=i+1; j<count && selS[j]-1<=selE[i]; j++); // empty statement
		selList.push({start: selS[i], end: selE[j-1]});
	}

	return selList;
}


// Initialization of rsseditor.xul window
function load()
{
	// setting the controller for the tree in order to make it respond to keyboard commands
	document.getElementById("tree").controllers.appendController(controller);

	// load prefs
	var prefCount = {};
	var prefArray = prefs.getChildList("", prefCount);
	for (var i = 0; i < prefCount.value; i++)
		gPrefListener.observe(prefs, "nsPref:changed", prefArray[i]); // mock pref change

	window.sdi = new sdiObject();

	// set listeners
	gPrefBranch.addObserver("", gPrefListener, false);
}


function unload()
{
	gPrefBranch.removeObserver("", gPrefListener);
}


// RSS Editor close confirmation - remember to use "return confirmClose()" for onClose attributes
function confirmClose()
{
	return confirmCloseFile();
}


function addListitem(item, endOfList)
{
	var list = document.getElementById("itemList");
	var e1 = document.createElement("treeitem");
	var e2 = document.createElement("treerow");
	const a = ["title", "pubDate", "category", "author"];
	for (var i in a)
		e2.appendChild(document.createElement("treecell")).setAttribute("label", item[a[i]]);
	e2.appendChild(document.createElement("treecell")).setAttribute("label", item.toSource());
	e1.appendChild(e2);
	if (selectedItem && !endOfList)
		list.insertBefore(e1, selectedItem.parentNode);
	else
		list.appendChild(e1);
	return e1;
}


// Events of items
function itemAdd(endOfList)
{
	var tree = document.getElementById("tree");
	var list = document.getElementById("itemList");
	var item = new RSSItem();
	item.title = localize("NewItem");
	item.guid = genGUID();

	try
	{
		updateItemCmdsDisabled = true;
		tree.view.selection.select(tree.view.getIndexOfItem(addListitem(item, endOfList)));
	}
	finally
	{
		updateItemCmdsDisabled = false;
		updateItemCmds();
	}
	if (tree.treeBoxObject.getLastVisibleRow < tree.currentIndex)
		tree.scrollToRow(tree.currentIndex - tree.treeBoxObject.getLastVisibleRow + tree.treeBoxObject.getFirstVisibleRow);
	dateUpdates();
}


function itemsDelete(confirm)
{
	if (confirm && openMsgBox(localize("ConfirmDeleteItemsMsg"), 33, localize("ConfirmDeleteItemsTitle")) != 1) return;

	var tree = document.getElementById("tree");
	var list = document.getElementById("itemList");

	var selList = treeSelection(tree);
	var selIndex = tree.currentIndex;
	var selectChanged = false;
	var count = selList.length;
	var range;

	try
	{
		updateItemCmdsDisabled = true;

		// deleting starts from the end of the list
		for (var i=count-1; i>=0; i--)
		{
			range = selList[i];
			if (selIndex >= range.start && selIndex <= range.end)
			{
				// move selection to the item immediately after the end of selection
				selIndex = range.start;
				// if no such item, select the item immediately before the selection
				if (range.end + 1 == tree.view.rowCount) selIndex--;
			}
			else if (!selectChanged && selIndex > range.end)
			{
				// actually moving selection in tree
				tree.view.selection.select(selIndex);
				selectChanged = true;
			}
			// delete items AFTER moving the selection
			for (var j=range.end; j>=range.start; j--)
				list.removeChild(tree.view.getItemAtIndex(j));
		}
		if (!selectChanged && tree.view.rowCount)
			tree.view.selection.select(selIndex > 0 ? selIndex : 0);
	}
	finally
	{
		updateItemCmdsDisabled = false;
		updateItemCmds();
	}
	channelLastBuildDateUpdate();
}


function itemsUp()
{
	var tree = document.getElementById("tree");
	var list = document.getElementById("itemList");

	var selList = treeSelection(tree);
	var count = selList.length;
	var range, item, refitem;


	if (selList[0].start <= 0) return
	try
	{
		itemSelectDisabled = true;
		updateItemCmdsDisabled = true;
		for (var i=0; i<count; i++)
		{
			range = selList[i];
			item = tree.view.getItemAtIndex(range.start - 1);
			if (range.end < tree.view.rowCount - 1)
			{
				refitem = tree.view.getItemAtIndex(range.end + 1);
				list.insertBefore(list.removeChild(item), refitem);
			}
			else
				list.appendChild(list.removeChild(item));
		}
	}
	finally
	{
		itemSelectDisabled = false;
		updateItemCmdsDisabled = false;
		updateItemCmds();
	}
	channelLastBuildDateUpdate();
}


function itemsDown()
{
	var tree = document.getElementById("tree");
	var list = document.getElementById("itemList");

	var selList = treeSelection(tree);
	var count = selList.length;
	var range, item, refitem;

	if (selList[count-1].end >= tree.view.rowCount-1) return;
	try
	{
		itemSelectDisabled = true;
		updateItemCmdsDisabled = true;
		for (var i=count-1; i>=0; i--)
		{
			range = selList[i];
			item = tree.view.getItemAtIndex(range.end + 1);
			refitem = tree.view.getItemAtIndex(range.start);
			list.insertBefore(list.removeChild(item), refitem);
		}
	}
	finally
	{
		itemSelectDisabled = false;
		updateItemCmdsDisabled = false;
		updateItemCmds();
	}
	channelLastBuildDateUpdate();
}


function itemSelect(skipUpdate)
{
	if (itemSelectDisabled) return;
	var tree = document.getElementById("tree");
	var row = tree.currentIndex;
	if (row < 0)
	{
		selectedItem = null;
		var item = new RSSItem();
	}
	else
	{
		var listitem = tree.view.getItemAtIndex(row).firstChild;

		skipUpdate || itemUpdate();
		selectedItem = listitem;

		try
		{
			item = eval(listitem.lastChild.getAttribute("label"));
		}
		catch (e) {}
	}
	setValue("itemTitle", item.title);
	setValue("itemLink", item.link);
	setValue("itemDesc", item.desc);
	setValue("itemAuthor", item.author);
	setValue("itemCategory", item.category);
	setValue("itemGuid", item.guid);
	document.getElementById("itemIsPermaLink").checked = item.isPermaLink;
	setValue("itemPubDate", item.pubDate);
	setValue("itemComments", item.comments);
	setValue("itemSource", item.source);
	setValue("itemSourceUrl", item.sourceurl);
	setValue("itemEnclosureUrl", item.enclosure.url);
	setValue("itemEnclosureLength", item.enclosure.length);
	setValue("itemEnclosureType", item.enclosure.type);
}


function itemInput(element)
{
	var tree = document.getElementById("tree");
	if (selectedItem == null) return;

	var itemNames = {"itemTitle":0, "itemPubDate":1, "itemCategory":2, "itemAuthor":3};
	var id = element.getAttribute("id");

	var colId = itemNames[id];
	if (id == "itemPubDate")
	{
		selectedItem.childNodes[colId].setAttribute("label", time(element.value));
		channelLastBuildDateUpdate();
	}
	else
	{
		colId != undefined && selectedItem.childNodes[colId].setAttribute("label", element.value);
		dateUpdates();
	}
}


function itemUpdate()
{

	if (selectedItem == null) return;
	var item = eval(selectedItem.lastChild.getAttribute("label"));

	item.title = getValue("itemTitle");
	item.link = getValue("itemLink");
	item.desc = getValue("itemDesc");
	item.author = getValue("itemAuthor");
	item.category = getValue("itemCategory");
	item.guid = getValue("itemGuid");
	item.isPermaLink = document.getElementById("itemIsPermaLink").checked;
	item.pubDate = time(getValue("itemPubDate"));
	item.comments = getValue("itemComments");
	item.source = getValue("itemSource");
	item.sourceurl = getValue("itemSourceUrl");
	item.enclosure.url = getValue("itemEnclosureUrl");
	item.enclosure.length = getValue("itemEnclosureLength");
	item.enclosure.type = getValue("itemEnclosureType");

	const a = ["title", "pubDate", "category", "author"];
	for (var i in a)
		selectedItem.childNodes[i].setAttribute("label", item[a[i]]);
	selectedItem.lastChild.setAttribute("label", item.toSource());
}


function itemGuidGen()
{
	if (!boolPref("guidReadOnly") || openMsgBox(localize("ConfirmChangeGUIDMsg"), 49, localize("ConfirmChangeGUIDTitle")) == 1)
	{
		setValue("itemGuid", genGUID());
		dateUpdates();
	}
}


function setDefault(arg)
{
	var target;
	switch (arg)
	{
		case "generator": target = "channelGenerator"; break;
		case "docs": target = "channelDocs"; break;
		default: throw("Invalid argument");
	}
	target = document.getElementById(target);
	if (target.value != gDefaults[arg])
	{
		target.value = gDefaults[arg];
		channelLastBuildDateUpdate();
	}
}


function publish()
{
	itemUpdate();
	window.openDialog("rsseditor_publish.xul", "_blank", "modal");
}


function options()
{
	// todo: if dialog is already open, use existing dialog instead of opening a new one
	window.openDialog("rsseditor_options.xul", "rsseditor_options");
}


function about()
{
	window.openDialog("rsseditor_about.xul", "rsseditor_about");
}